/* api-wrapper.c - Power Station hi-level functions implementation
 * based on PhyMod Library
 * Copyright (c) 2000 David A. Bartold
 * Copyright (c) 2004, 2020 Yury G. Aliaev
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>

#include "api-wrapper.h"

/* Now len means _maximal_ lenght if the given attenuation will not be reached;
   for disabling stopping at given attenuation, use attenuation = 0.0.
   Attenuation is given in dB, att = 60.0 means render will be stopped after
   the mean amplitude reach the value of -60 dB. */
gsize
ps_metal_obj_render(PSPhysMod *model, gint rate,
	gsize len, gdouble * samples,
	gdouble att, gpointer userdata)
{
    gsize i, real_len;
    gdouble maxvol;
    gdouble stasis, damp;
    gdouble sample, hipass, hipass_coeff, lowpass_coeff, lowpass, maxamp;

    gdouble curr_att = 0.0;
    gfloat per_cent = 0.0;

    PSMetalObj *obj;

	model->stop = FALSE;

	switch(model->obj_type) {
	case PS_OBJECT_TUBE:
		obj = ps_metal_obj_new_tube(model->height, model->circum, model->tenseness);
		break;
	case PS_OBJECT_ROD:
		obj = ps_metal_obj_new_rod(model->length, model->tenseness);
		break;
	case PS_OBJECT_PLANE:
		obj = ps_metal_obj_new_plane(model->plane_length, model->plane_width, model->tenseness);
		break;
	default:
		g_assert_not_reached();
	}

    if (model->percent_callback != NULL)
        model->percent_callback(0.0, userdata);

    if (model->actuation) {
	stasis = obj->nodes[model->outnode]->pos.z;
	obj->nodes[model->innode]->pos.z += model->velocity;
    } else {
	stasis = obj->nodes[model->outnode]->pos.x;
	obj->nodes[model->innode]->pos.x += model->velocity;
    }

    hipass = lowpass = maxamp = 0.0;
    hipass_coeff = pow(0.5, 5.0 / rate);
    lowpass_coeff = 1 - 20.0 / rate;	/* 50 ms integrator */
    damp = pow(0.5, 1.0 / (model->damping * rate));

    maxvol = 0.0;
    for (i = 0; i < len && !model->stop; i++) {
	ps_metal_obj_perturb(obj, model->speed * 44100.0 / (gdouble)rate, damp);

	if (model->actuation)
	    sample = obj->nodes[model->outnode]->pos.z - stasis;
	else
	    sample = obj->nodes[model->outnode]->pos.x - stasis;

	hipass = hipass_coeff * hipass + (1.0 - hipass_coeff) * sample;
	samples[i] = sample - hipass;

	if (fabs(samples[i]) > maxvol)
	    maxvol = fabs(samples[i]);

	lowpass =
	    lowpass_coeff * lowpass + (1.0 -
				       lowpass_coeff) * fabs(samples[i]);
	if (maxamp < lowpass)
	    maxamp = lowpass;

	if (att < 0) {
	    if (maxamp > 0.0)
		curr_att = 20 * log10(lowpass / maxamp);

	    if (model->percent_callback != NULL) {
		    gfloat p1, p2, p3;

		    p1 = ((gfloat) i) / len;
		    p2 = curr_att / att;
		    p3 = MAX(p1, p2);
		    if (p3 - per_cent > 0.01) {
		        per_cent = p3;
		        model->percent_callback(p3, userdata);
		    }
		}

	    if (curr_att <= att)
		break;
	} else if (model->percent_callback != NULL) {
            gfloat p = ((gfloat) i) / len;

            if (p - per_cent > 0.01) {
                per_cent = p;
                model->percent_callback(p, userdata);
            }
        }
    }
    ps_metal_obj_free(obj);
    real_len = i;

    if (maxvol > 0.0) {
        maxvol = 1.0 / maxvol;
        for (i = 0; i < real_len; i++)
	    samples[i] *= maxvol;
    }

    return model->stop ? 0 : real_len;
}
