/*  Power Station Industrializer
 *  Copyright (c) 2000 David A. Bartold
 *  Copyright (c) 2020 Yury G. Aliaev
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <string.h>
#ifdef HAVE_OPENGL
#  include <gtk/gtkgl.h>
#endif

#include "interface.h"
#include "main.h"
#include "utils.h"

#ifdef DRIVER_ALSA
#include "alsa.h"
#endif

#ifdef DRIVER_PULSE
#include "pulse.h"
#endif

#ifdef DRIVER_JACK
#include "jack.h"
#endif

typedef enum
{
    CONF_INT,
    CONF_BOOL,
    CONF_STRING,
    CONF_ARRAY_S
} ConfType;

struct config_array
{
    const gchar *path;
    const gchar *name;
    ConfType type;
    union {
        gpointer p;
        gint i;
        gboolean b;
    } value;
};

drv *driver = NULL;
GSList *driver_list = NULL;
xmlpContext *config;

static const gchar conf_ver[] = "0.1.0";

static guint current_driver;

static struct config_array *conf_array;
static gint c_index = 0, c_num = 256;

guint psi_get_current_driver(void)
{
    return current_driver;
}

void psi_driver_errmessage(int err_no)
{
    static GtkWidget *dialog = NULL;
    gchar *buf = g_strdup_printf("%s:\n%s", _("Sound driver error"), driver->err(err_no));

    gui_error_dialog(&dialog, buf);
}

void psi_set_driver(const guint drv, const gint prate)
{
    int err;

    /* Some drivers may have no activate() and/or deactivate() methods */
    if (driver && driver->deactivate)
        driver->deactivate();
    driver = g_slist_nth_data(driver_list, drv);
    current_driver = drv;
    if (driver && driver->activate)
    {
        if ((err = driver->activate(prate)) < 0) {
            psi_driver_errmessage(err);
            driver = NULL;
        }
    }
}

void
conf_change_path(const gint tag, const gchar *path)
{
    conf_array[tag].path = path;
}

void
conf_rename(const gint tag, const gchar *name)
{
    conf_array[tag].name = name;
}

static void
check_array_size(gint in)
{
    if(in >= c_num) {
        c_num += c_num >> 1;
        conf_array = g_renew(struct config_array, conf_array, c_num);
    }
}

gint
conf_subscribe_bool(const gchar *path, const gchar *name, const gboolean deflt)
{
    gboolean val;
    gint tag = c_index;

    g_assert(config != NULL);

    val = xmlp_get_boolean_default(config, path, name, deflt);
    conf_array[c_index].path = path;
    conf_array[c_index].name = name;
    conf_array[c_index].type = CONF_BOOL;
    conf_array[c_index].value.b = val;

    check_array_size(++c_index);

    return tag;
}

gboolean
conf_get_bool(gint tag)
{
    return conf_array[tag].value.b;
}

void
conf_set_bool(gint tag, gboolean val)
{
    conf_array[tag].value.b = val;
}

gint
conf_subscribe_int(const gchar *path, const gchar *name, const gint deflt)
{
    gint val;
    gint tag = c_index;

    g_assert(config != NULL);

    val = xmlp_get_int_default(config, path, name, deflt);
    conf_array[c_index].path = path;
    conf_array[c_index].name = name;
    conf_array[c_index].type = CONF_INT;
    conf_array[c_index].value.i = val;

    check_array_size(++c_index);

    return tag;
}

gint
conf_get_int(gint tag)
{
    return conf_array[tag].value.i;
}

void
conf_set_int(gint tag, gint val)
{
    conf_array[tag].value.i = val;
}

gint
conf_subscribe_string(const gchar *path, const gchar *name, const gchar *deflt)
{
    gchar *val;
    gint tag = c_index;

    g_assert(config != NULL);

    val = xmlp_get_string(config, path, name);
    if(!val && deflt)
        val = g_strdup(deflt);
    conf_array[c_index].path = path;
    conf_array[c_index].name = name;
    conf_array[c_index].type = CONF_STRING;
    conf_array[c_index].value.p = val;

    check_array_size(++c_index);

    return tag;
}

gchar*
conf_get_string(gint tag)
{
    return conf_array[tag].value.p;
}

void
conf_set_string(gint tag, gchar *val)
{
    conf_array[tag].value.p = val;
}

gint
conf_subscribe_string_array(const gchar *path, const gchar *name, const gchar *deflt)
{
    GSList *val;
    gint tag = c_index;

    g_assert(config != NULL);

    val = xmlp_get_string_array(config, path, name);
    if(!val && deflt)
        val = g_slist_append(NULL, xmlStrdup((const xmlChar*)deflt));
    conf_array[c_index].path = path;
    conf_array[c_index].name = name;
    conf_array[c_index].type = CONF_ARRAY_S;
    conf_array[c_index].value.p = val;

    check_array_size(++c_index);

    return tag;
}

GSList*
conf_get_string_array(gint tag)
{
    return conf_array[tag].value.p;
}

void
conf_set_string_array(gint tag, GSList *val)
{
    conf_array[tag].value.p = val;
}

int main(int argc, char *argv[])
{
    GSList *currd;
    drv *d;
    struct stat		st;
    struct passwd	*pw;
    gchar		*confdir, *conffile, *current_driver_string;
    guint i;

#ifdef ENABLE_NLS
    bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
    bind_textdomain_codeset(PACKAGE,"UTF-8");
    textdomain(PACKAGE);
#endif

    gtk_init(&argc, &argv);
#ifdef HAVE_OPENGL
    gtk_gl_init (&argc, &argv);
#endif

#ifdef DRIVER_PULSE
    driver_list = g_slist_append(driver_list, &driver_pulse);
#endif
#ifdef DRIVER_JACK
    driver_list = g_slist_append(driver_list, &driver_jack);
#endif
#ifdef DRIVER_ALSA
    driver_list = g_slist_append(driver_list, &driver_alsa);
#endif

    pw = getpwuid(getuid());
    confdir = g_strconcat(pw->pw_dir, "/."PACKAGE, NULL);
    conffile = g_strconcat(confdir, "/config", NULL);

    if(stat(confdir, &st) < 0)
	mkdir(confdir, S_IRUSR | S_IWUSR | S_IXUSR);

    if(!(config = xmlp_get_doc(conffile, "psiconfig", conf_ver)))
	config = xmlp_new_doc(conffile, "psiconfig", conf_ver);
    g_free(confdir);
    conf_array = g_new(struct config_array, c_num);

    for(currd = driver_list; currd; currd = currd->next) {
        d = currd->data;

        if(d->loadsettings)
            d->loadsettings();
    }
    current_driver = 0;
    if((current_driver_string = xmlp_get_string(config, "driver/", "current"))) {
        for(i = 0; i < g_slist_length(driver_list); i++) {
            d = g_slist_nth_data(driver_list, i);

            if(!g_ascii_strcasecmp(d->description, current_driver_string))
                current_driver = i;
        }
        xmlp_free_string(current_driver_string);
    }

    gui_create_AppWindow(current_driver);
    gtk_main();

    if(config->old_mode) {
        /* We just recreate the config file if the existing was from an old version
           of PSI. */
        xmlp_free(config);
        config = xmlp_new_doc(conffile, "psiconfig", conf_ver);
    }
    g_free(conffile);

    for(i = 0; i < c_index; i++)
        switch(conf_array[i].type){
        case CONF_INT:
            xmlp_set_int(config, conf_array[i].path,
                conf_array[i].name, conf_array[i].value.i);
            break;
        case CONF_BOOL:
            xmlp_set_boolean(config, conf_array[i].path,
                conf_array[i].name, conf_array[i].value.b);
            break;
        case CONF_STRING:
            xmlp_set_string(config, conf_array[i].path,
                conf_array[i].name, conf_array[i].value.p);
            break;
        case CONF_ARRAY_S:
            xmlp_set_string_array(config, conf_array[i].path,
                conf_array[i].name, conf_array[i].value.p);
            break;
        }

    if(driver)
        xmlp_set_string(config, "driver/", "current", driver->description);
    xmlp_sync(config);

    if (driver && driver->deactivate)
        driver->deactivate();

    return 0;
}
