/* xml-parser.c - xml files manipulating functions
 * Copyright (c) 2000 David A. Bartold
 * Copyright (c) 2005-2020 Yury G. Aliaev
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <glib/gprintf.h>

#include "xml-parser.h"

static gchar buf[256];
static const xmlChar ver_name[] = "version";
static const xmlChar val_name[] = "value";

xmlpContext*
xmlp_new_doc (const gchar *docname, const gchar *keyword, const gchar *version)
{
    xmlpContext *cntxt;
    xmlNodePtr	cur;

    cntxt = g_new(xmlpContext, 1);
    cntxt->name = g_strdup(docname);
    cntxt->type = g_strdup(keyword);
    cntxt->old_mode = FALSE;

    cntxt->doc = xmlNewDoc((const xmlChar*)"1.0");
    cur = xmlNewDocNode(cntxt->doc, NULL, (const xmlChar*)keyword, NULL);
    xmlDocSetRootElement(cntxt->doc, cur);
    xmlNewProp(cur, (const xmlChar*)"version", (const xmlChar*)version);

    cntxt->context = xmlXPathNewContext(cntxt->doc);
    return cntxt;
}

xmlpContext*
xmlp_get_doc (const gchar *docname, const gchar *keyword, const gchar *version)
{
    xmlpContext *cntxt;
    xmlNodePtr cur;
    xmlDocPtr doc;
    xmlChar *prop;

    xmlKeepBlanksDefault(0);

    if((doc = xmlParseFile(docname)) == NULL)
	return NULL;

    if(!(cur = xmlDocGetRootElement(doc))) {
	xmlFreeDoc(doc);
	return NULL;
    }

    /* Incorrect (mismatching) type */
    if(xmlStrcasecmp(cur->name, (const xmlChar*)keyword)){
	xmlFreeDoc(doc);
	return NULL;
    }

    cntxt = g_new(xmlpContext, 1);
    cntxt->doc = doc;

    prop = xmlGetProp(cur, ver_name);
    if (!prop) {
        cntxt->old_mode = TRUE;
        xmlNewProp(cur, ver_name, (const xmlChar*)version);
    } else {
        cntxt->old_mode = FALSE;
        xmlSetProp(cur, ver_name, (const xmlChar*)version);
        xmlFree(prop);
    }

    cntxt->name = g_strdup(docname);
    cntxt->type = g_strdup(keyword);
    cntxt->context = xmlXPathNewContext(cntxt->doc);
    return cntxt;
}

void
xmlp_free (xmlpContext *cntxt)
{
    xmlXPathFreeContext(cntxt->context);
    xmlFreeDoc(cntxt->doc);
    g_free(cntxt->name);
    g_free(cntxt->type);
    g_free(cntxt);
    xmlCleanupParser();
}

void
xmlp_free_string (gchar *str)
{
    xmlFree((xmlChar *)str);
}

/* Required g_strfreev() on non-NULL result */
GSList*
xmlp_get_array (xmlpContext *cntxt, const gchar *xpath, const gchar *subtype, const gchar *name)
{
    xmlXPathObjectPtr	result;
    GSList *value = NULL;
    gchar		*path;

    path = g_strconcat("//", xpath,
        "array[attribute::name=\"", name, "\"]/", subtype, NULL);
    result = xmlXPathEvalExpression((xmlChar *)path, cntxt->context);
    g_free(path);

    if(!xmlXPathNodeSetIsEmpty(result->nodesetval)) {
        guint i;
        xmlNodeSetPtr nodeset = result->nodesetval;

        for (i = 0; i < nodeset->nodeNr; i++) {
            xmlChar *line = xmlGetProp(nodeset->nodeTab[i], val_name);

            value = g_slist_append(value, line);
        }
    }

    xmlXPathFreeObject (result);
    return value;
}

/*
void
xmlp_free_array (GSList *array)
{
    GSList *cur = array;
    for(; cur; cur = cur->next)
        xmlFree(cur->data);
    g_slist_free(array);
}
*/

/* These functions return only the first value occurens in the case of multiple */
xmlChar*
xmlp_get (xmlpContext *cntxt, const gchar *xpath, const gchar *type, const gchar *name)
{
    xmlXPathObjectPtr	result;
    xmlChar		*value = NULL;
    gchar		*path;

    path = g_strconcat("//", xpath, type, "[attribute::name=\"", name, "\"]", NULL);
    result = xmlXPathEvalExpression((xmlChar *)path, cntxt->context);
    g_free(path);

    if(!xmlXPathNodeSetIsEmpty(result->nodesetval)) {
        xmlNodeSetPtr nodeset = result->nodesetval;

        value = cntxt->old_mode ?
            xmlNodeListGetString(cntxt->doc, nodeset->nodeTab[0]->xmlChildrenNode, 1) :
            xmlGetProp(nodeset->nodeTab[0], val_name);
    }

    xmlXPathFreeObject (result);
    return value;
}

gchar*
xmlp_get_string (xmlpContext *cntxt, const gchar *xpath, const gchar *name)
{
    return (gchar *)xmlp_get(cntxt, xpath, "string", name);
}

gchar*
xmlp_get_string_default (xmlpContext *cntxt, const gchar *xpath, const gchar *name, const gchar *deflt)
{
    gchar	*str;
    
    if((str = xmlp_get_string(cntxt, xpath, name)))
	return str;
    else {
	str = g_strdup(deflt);
	return str;
    }
}
/*
gchar*
xmlp_get_set_string_default (xmlpContext *cntxt, const gchar *xpath, const gchar *name, const gchar *deflt)
{
    gchar	*str;
    
    if((str = xmlp_get_string(cntxt, xpath, name)))
	return str;
    else {
	str = g_strdup(deflt);
	xmlp_set_string(cntxt, xpath, name, str);
	return str;
    }
}
*/
gboolean
xmlp_get_int (xmlpContext *cntxt, const gchar *xpath, const gchar *name, gint *value)
{
    xmlChar	*strvalue;
    
    if((strvalue = xmlp_get(cntxt, xpath, "int", name))) {
	*value = atoi((char *)strvalue);
	xmlFree(strvalue);
	return TRUE;
    }
    
    return FALSE;
}

gint
xmlp_get_int_default (xmlpContext *cntxt, const gchar *xpath, const gchar *name, const gint value)
{
    gint	result = value;

    xmlp_get_int (cntxt, xpath, name, &result);
    return result;
}

gboolean
xmlp_get_double (xmlpContext *cntxt, const gchar *xpath, const gchar *name, gdouble *value)
{
    xmlChar	*strvalue;

    if((strvalue = xmlp_get(cntxt, xpath, "double", name))) {
        *value = cntxt->old_mode ? g_strtod((gchar *)strvalue, NULL)
                                 : g_ascii_strtod((gchar *)strvalue, NULL);
        xmlFree(strvalue);
        return TRUE;
    }

    return FALSE;
}

gdouble
xmlp_get_double_default (xmlpContext *cntxt, const gchar *xpath, const gchar *name, const gdouble value)
{
    gdouble	result = value;

    xmlp_get_double (cntxt, xpath, name, &result);
    return result;
}

gboolean
xmlp_get_boolean (xmlpContext *cntxt, const gchar *xpath, const gchar *name, gboolean *value)
{
    xmlChar	*strvalue;
    
    if((strvalue = xmlp_get(cntxt, xpath, "bool", name))) {
	if(!xmlStrcasecmp(strvalue, (const xmlChar*)"true"))
	    *value = TRUE;
	else
	    *value = FALSE;
	xmlFree(strvalue);
	return TRUE;
    }
    
    return FALSE;
}

gboolean
xmlp_get_boolean_default (xmlpContext *cntxt, const gchar *xpath, const gchar *name, const gboolean value)
{
    gboolean	result = value;

    xmlp_get_boolean (cntxt, xpath, name, &result);
    return result;
}

static xmlNodePtr
lookup_node (xmlNodePtr cur, const xmlChar *name, const xmlChar *aname, const xmlChar *avalue)
{
    for (;cur;cur = cur->next) {
        if (!xmlStrcasecmp(cur->name, name)) {
            if (aname) {
                xmlChar *prop = xmlGetProp(cur, aname);

                if (!xmlStrcasecmp(prop, avalue)) {
                    xmlFree(prop);
                    break;
                }
                xmlFree(prop);
            } else
                break;
        }
    }

    return cur;
}

static xmlNodePtr
xmlp_set_common (xmlpContext *cntxt, const gchar *xpath, const gchar *type, const gchar *name)
{
    gchar *path;
    gchar **subpath;
    xmlNodePtr cur;
    guint i;
    gboolean exists = TRUE;

    path = g_strconcat(xpath, type, "[attribute::name=\"", name, "\"]", NULL);
    cur = xmlDocGetRootElement(cntxt->doc);

    subpath = g_strsplit(path, "/", -1);
    g_assert(subpath);

    for (i = 0; subpath[i]; i++) {
        gchar *aname = NULL, **aname_aval = NULL; /* NULLs are just to avoid compiler warnings */
        gchar **name_attr = g_strsplit(subpath[i], "[attribute::", 2);

        if (name_attr[1]) {
            aname_aval = g_strsplit(name_attr[1], "\"", 3);
            aname = g_strndup(aname_aval[0], strlen(aname_aval[0]) - 1); /* Remove the trailing "=" */
        }
        if (exists) {
            xmlNodePtr tmp = lookup_node(cur->xmlChildrenNode,
                (const xmlChar*)name_attr[0],
                name_attr[1] ? (const xmlChar*)aname : NULL,
                name_attr[1] ? (const xmlChar*)aname_aval[1] : NULL);
            if (!tmp)
                exists = FALSE;
            else
                cur = tmp;
        }
        if (!exists) {
            cur = xmlNewTextChild(cur, NULL, (const xmlChar*)name_attr[0], NULL);
            if (name_attr[1])
                xmlNewProp(cur, (const xmlChar*)aname, (const xmlChar*)aname_aval[1]);
        }
        if (name_attr[1]) {
            g_strfreev(aname_aval);
            g_free(aname);
        }
        g_strfreev(name_attr);
    }
    g_strfreev(subpath);

    return cur;
}

void
xmlp_set (xmlpContext *cntxt, const gchar *xpath, const gchar *type, const gchar *name, const gchar *value)
{
    xmlNodePtr cur = xmlp_set_common(cntxt, xpath, type, name);
    xmlChar *prop;

    prop = xmlGetProp(cur, val_name);
    if (prop) {
        xmlSetProp(cur, val_name, (const xmlChar*)value);
        xmlFree(prop);
    } else
        xmlNewProp(cur, val_name, (const xmlChar*)value);
}

void
xmlp_set_array (xmlpContext *cntxt, const gchar *xpath, const gchar *type, const gchar *name, GSList *values)
{
    xmlNodePtr cur = xmlp_set_common(cntxt, xpath, "array", name);
    xmlNodePtr children = cur->xmlChildrenNode;

    /* We just remove everything inside <array>...</array> and create new contents */
    for(; children;){
        xmlNodePtr tmp = children->next;
        xmlUnlinkNode(children);
        children = tmp;
    }
    for(; values; values = values->next) {
        xmlNodePtr child = xmlNewTextChild(cur, NULL, (const xmlChar*)type, NULL);
        xmlNewProp(child, val_name, (const xmlChar*)values->data);
    }
}

void
xmlp_set_string (xmlpContext *cntxt, const gchar *xpath, const gchar *name, const gchar *value)
{
    xmlp_set(cntxt, xpath, "string", name, value);
}

void
xmlp_set_int (xmlpContext *cntxt, const gchar *xpath, const gchar *name, gint value)
{
    g_sprintf(buf, "%i", value);
    xmlp_set(cntxt, xpath, "int", name, buf);
}

void
xmlp_set_double (xmlpContext *cntxt, const gchar *xpath, const gchar *name, gdouble value)
{
    g_ascii_dtostr(buf, sizeof(buf), value);
    xmlp_set(cntxt, xpath, "double", name, buf);
}

void
xmlp_set_boolean (xmlpContext *cntxt, const gchar *xpath, const gchar *name, gboolean value)
{
    xmlp_set(cntxt, xpath, "bool", name, value ? "TRUE" : "FALSE");
}

void
xmlp_sync (xmlpContext *cntxt)
{
    xmlSaveFormatFile(cntxt->name, cntxt->doc, 1);
}
